# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause


def raise_error_unsupported_categorical_option(
    option_name, option_value, layer_type, layer_name
):
    """
    Raise an error if an option is not supported.
    """
    raise RuntimeError(
        "Unsupported option %s=%s in layer %s(%s)"
        % (option_name, option_value, layer_type, layer_name)
    )


def raise_error_unsupported_option(option, layer_type, layer_name):
    """
    Raise an error if an option is not supported.
    """
    raise RuntimeError(
        "Unsupported option =%s in layer %s(%s)" % (option, layer_type, layer_name)
    )


def raise_error_unsupported_scenario(message, layer_type, layer_name):
    """
    Raise an error if an scenario is not supported.
    """
    raise RuntimeError(
        "Unsupported scenario '%s' in layer %s(%s)" % (message, layer_type, layer_name)
    )



from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _

from typing import Dict, Text, Any, Callable
from coremltools.models.neural_network import NeuralNetworkBuilder  # type: ignore
from ._graph import Node, Graph


class ErrorHandling(object):
    """
  To handle errors and addition of custom layers
  """

    def __init__(
        self,
        add_custom_layers=False,  # type: bool
        custom_conversion_functions=dict(),  # type: Dict[Text, Any]
        custom_layer_nodes=[],  # type : List[Node]
    ):
        # type: (...) -> None
        self.add_custom_layers = add_custom_layers
        self.custom_conversion_functions = custom_conversion_functions
        self.custom_layer_nodes = custom_layer_nodes

        self.rerun_suggestion = (
            "\n Please try converting with higher minimum_ios_deployment_target.\n"
            "You can also provide custom function/layer to convert the model."
        )

    def unsupported_op(
        self, node,  # type: Node
    ):
        # type: (...) -> Callable[[Any, Node, Graph, ErrorHandling], None]
        """
      Either raise an error for an unsupported op type or return custom layer add function
      """
        if self.add_custom_layers:
            from ._operators import _convert_custom

            return _convert_custom
        else:
            raise TypeError(
                "ONNX node of type {} is not supported. {}\n".format(
                    node.op_type, self.rerun_suggestion
                )
            )

    def unsupported_op_configuration(
        self,
        builder,  # type: NeuralNetworkBuilder
        node,  # type: Node
        graph,  # type: Graph
        err_message,  # type: Text
    ):
        # type: (...) -> None
        """
      Either raise an error for an unsupported attribute or add a custom layer.
      """
        if self.add_custom_layers:
            from ._operators import _convert_custom

            _convert_custom(builder, node, graph, self)
        else:
            raise TypeError(
                "Error while converting op of type: {}. Error message: {} {}\n".format(
                    node.op_type, err_message, self.rerun_suggestion
                )
            )

    def missing_initializer(
        self,
        node,  # type: Node
        err_message,  # type: Text
    ):
        # type: (...) -> None
        """
      Missing initializer error
      """
        raise ValueError(
            "Missing initializer error in op of type {}, with input name = {}, "
            "output name = {}. Error message: {} {}\n".format(
                node.op_type,
                node.inputs[0],
                node.outputs[0],
                err_message,
                self.rerun_suggestion,
            )
        )

    def unsupported_feature_warning(
        self,
        node,  # type: Node
        warn_message,  # type: Text
    ):
        # type: (...) -> None
        """
      Unsupported feature warning
      """
        print(
            "Warning: Unsupported Feature in op of type {}, with input name = {}, "
            "output name = {}. Warning message: {}\n".format(
                node.op_type, node.inputs[0], node.outputs[0], warn_message
            )
        )



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ..._deps import _HAS_LIBSVM


def load_model(model_path):
    """Load a libsvm model from a path on disk.

    This currently supports:
      * C-SVC
      * NU-SVC
      * Epsilon-SVR
      * NU-SVR

    Parameters
    ----------
    model_path: str
        Path on disk where the libsvm model representation is.

    Returns
    -------
    model: libsvm_model
        A model of the libsvm format.
    """
    if not (_HAS_LIBSVM):
        raise RuntimeError("libsvm not found. libsvm conversion API is disabled.")

    from svmutil import svm_load_model  # From libsvm
    import os

    if not os.path.exists(model_path):
        raise IOError("Expected a valid file path. %s does not exist" % model_path)
    return svm_load_model(model_path)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: LinkedModel.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import Parameters_pb2 as Parameters__pb2
try:
  DataStructures__pb2 = Parameters__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = Parameters__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes_pb2

from .Parameters_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='LinkedModel.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x11LinkedModel.proto\x12\x14\x43oreML.Specification\x1a\x10Parameters.proto\"[\n\x0bLinkedModel\x12@\n\x0flinkedModelFile\x18\x01 \x01(\x0b\x32%.CoreML.Specification.LinkedModelFileH\x00\x42\n\n\x08LinkType\"\x9b\x01\n\x0fLinkedModelFile\x12\x42\n\x13linkedModelFileName\x18\x01 \x01(\x0b\x32%.CoreML.Specification.StringParameter\x12\x44\n\x15linkedModelSearchPath\x18\x02 \x01(\x0b\x32%.CoreML.Specification.StringParameterB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[Parameters__pb2.DESCRIPTOR,],
  public_dependencies=[Parameters__pb2.DESCRIPTOR,])




_LINKEDMODEL = _descriptor.Descriptor(
  name='LinkedModel',
  full_name='CoreML.Specification.LinkedModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='linkedModelFile', full_name='CoreML.Specification.LinkedModel.linkedModelFile', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='LinkType', full_name='CoreML.Specification.LinkedModel.LinkType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=61,
  serialized_end=152,
)


_LINKEDMODELFILE = _descriptor.Descriptor(
  name='LinkedModelFile',
  full_name='CoreML.Specification.LinkedModelFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='linkedModelFileName', full_name='CoreML.Specification.LinkedModelFile.linkedModelFileName', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='linkedModelSearchPath', full_name='CoreML.Specification.LinkedModelFile.linkedModelSearchPath', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=310,
)

_LINKEDMODEL.fields_by_name['linkedModelFile'].message_type = _LINKEDMODELFILE
_LINKEDMODEL.oneofs_by_name['LinkType'].fields.append(
  _LINKEDMODEL.fields_by_name['linkedModelFile'])
_LINKEDMODEL.fields_by_name['linkedModelFile'].containing_oneof = _LINKEDMODEL.oneofs_by_name['LinkType']
_LINKEDMODELFILE.fields_by_name['linkedModelFileName'].message_type = Parameters__pb2._STRINGPARAMETER
_LINKEDMODELFILE.fields_by_name['linkedModelSearchPath'].message_type = Parameters__pb2._STRINGPARAMETER
DESCRIPTOR.message_types_by_name['LinkedModel'] = _LINKEDMODEL
DESCRIPTOR.message_types_by_name['LinkedModelFile'] = _LINKEDMODELFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LinkedModel = _reflection.GeneratedProtocolMessageType('LinkedModel', (_message.Message,), dict(
  DESCRIPTOR = _LINKEDMODEL,
  __module__ = 'LinkedModel_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.LinkedModel)
  ))
_sym_db.RegisterMessage(LinkedModel)

LinkedModelFile = _reflection.GeneratedProtocolMessageType('LinkedModelFile', (_message.Message,), dict(
  DESCRIPTOR = _LINKEDMODELFILE,
  __module__ = 'LinkedModel_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.LinkedModelFile)
  ))
_sym_db.RegisterMessage(LinkedModelFile)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



import os


def _shape_notation(int_shape):
    X = ["S", "B", "C", "H", "W"]
    return [X[i] for i in int_shape]


def plot_graph(graph, graph_img_path="graph.png", show_coreml_mapped_shapes=False):
    """
    Plot graph using pydot

    It works in two steps:
    1. Add nodes to pydot
    2. connect nodes added in pydot

    :param graph
    :return: writes down a png/pdf file using dot
    """

    try:
        # pydot-ng is a fork of pydot that is better maintained.
        import pydot_ng as pydot  # type: ignore
    except:
        # pydotplus is an improved version of pydot
        try:
            import pydotplus as pydot  # type: ignore
        except:
            # Fall back on pydot if necessary.
            try:
                import pydot  # type: ignore
            except:
                return None

    dot = pydot.Dot()
    dot.set("rankdir", "TB")
    dot.set("concentrate", True)
    dot.set_node_defaults(shape="record")

    # Add nodes corresponding to graph inputs
    graph_inputs = []
    for input_ in graph.inputs:
        if show_coreml_mapped_shapes:
            if input_[0] in graph.onnx_coreml_shape_mapping:
                shape = tuple(
                    _shape_notation(graph.onnx_coreml_shape_mapping[input_[0]])
                )
            else:
                shape = "NA, "
        else:
            shape = tuple(input_[2])
        label = "%s\n|{|%s}|{{%s}|{%s}}" % ("Input", input_[0], "", str(shape))
        pydot_node = pydot.Node(input_[0], label=label)
        dot.add_node(pydot_node)
        graph_inputs.append(input_[0])

    # Traverse graph and add nodes to pydot
    for node in graph.nodes:
        inputlabels = ""
        for input_ in node.inputs:
            if show_coreml_mapped_shapes:
                if input_ in graph.onnx_coreml_shape_mapping:
                    inputlabels += (
                        str(
                            tuple(
                                _shape_notation(graph.onnx_coreml_shape_mapping[input_])
                            )
                        )
                        + ", "
                    )
                else:
                    inputlabels += "NA, "
            else:
                if input_ in graph.shape_dict:
                    inputlabels += str(tuple(graph.shape_dict[input_])) + ", "
                else:
                    inputlabels += "NA, "
        outputlabels = ""
        for output_ in node.outputs:
            if show_coreml_mapped_shapes:
                if output_ in graph.onnx_coreml_shape_mapping:
                    outputlabels += (
                        str(
                            tuple(
                                _shape_notation(
                                    graph.onnx_coreml_shape_mapping[output_]
                                )
                            )
                        )
                        + ", "
                    )
                else:
                    outputlabels += "NA, "
            else:
                if output_ in graph.shape_dict:
                    outputlabels += str(tuple(graph.shape_dict[output_])) + ", "
                else:
                    outputlabels += "NA, "
        output_names = ", ".join([output_ for output_ in node.outputs])
        input_names = ", ".join([input_ for input_ in node.inputs])
        label = "%s\n|{{%s}|{%s}}|{{%s}|{%s}}" % (
            node.op_type,
            input_names,
            output_names,
            inputlabels,
            outputlabels,
        )
        pydot_node = pydot.Node(node.name, label=label)
        dot.add_node(pydot_node)

    # add edges
    for node in graph.nodes:
        for child in node.children:
            # add edge in pydot
            dot.add_edge(pydot.Edge(node.name, child.name))
        for input_ in node.inputs:
            if input_ in graph_inputs:
                dot.add_edge(pydot.Edge(input_, node.name))

    # write out the image file
    _, extension = os.path.splitext(graph_img_path)
    if not extension:
        extension = "pdf"
    else:
        extension = extension[1:]
    dot.write(graph_img_path, format=extension)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil import SYMBOL, VALUE
from coremltools.converters.mil.mil.types.symbolic import is_compatible_symbolic_vector
from ._op_reqs import *
import numbers


@register_op(doc_str="")
class gather(Operation):
    """
    Gather slices from input ``x`` along dimension ``axis`` according to ``indices``,
    similar to `tf.gather <https://www.tensorflow.org/api_docs/python/tf/gather>`_.
    
    * If ``indices`` is scalar (0-D):
    
    .. math::
       output[p_0, ..., p_{axis-1}, ~~~~~~~~~~~~~~~~~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}] =
    .. math::
       x[p_0, ..., p_{axis-1}, ~~~~~~~~~ indices, ~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}]
    
    Where ``rank(x)`` is the rank of ``x``. The ``output`` has rank ``rank(x) - 1``.
    
    * If ``indices`` is 1-D tensor:
    
    .. math::
       output[p_0, ..., p_{axis-1}, ~~~~~~~~~~~~~ i, ~~~~~~~~~~~~~ p_{axis+1}, ..., p_{rank(*D)-1}] =
    .. math::
       x[p_0, ..., p_{axis-1}, ~~~~~~~~ indices[i], ~~~~~~~~ p_{axis+1}, ..., p_{rank(*D)-1}]
    
    The output has rank ``rank(x)``.
    
    * In general:
    
    .. math::
       output[p_0, ..., p_{axis-1}, ~~~~~~~~ i_0, ..., i_{M-1}, ~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}] =
    .. math::
       x[p_0, ..., p_{axis-1}, ~~~~~~~ indices[i_0, ..., i_{M-1}], ~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}]
    
    Where ``M = rank(x)``.
    
    Parameters
    ----------
    x: tensor<\*D,T> (Required)
    indices: tensor<\*N,i32> (Required)
        * Indices values may be negative. More precisely, ``-D[axis]<= v < D[axis]`` for ``v`` in ``indices``.
    axis: const i32 (Optional. Default=``0``)
        * Negative axis is supported.
    
    Returns
    -------
    tensor<\*K,T>
        * Where ``K = D[:axis] + N + D[axis+1:]``.
    
    Attributes
    ----------
    T: fp32
    
    References
    ----------
    See `tf.gather <https://www.tensorflow.org/api_docs/python/tf/gather>`_.
    
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        indices=IntInputType(),
        axis=IntInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axis=0,
            )

    def __init__(self, **kwargs):
        super(gather, self).__init__(**kwargs)

    @precondition(allow=VALUE | SYMBOL)
    def value_inference(self):
        x = self.x.sym_val
        indices = self.indices.val
        if indices is None:
            # only allow x to be symbolic. indices cannot.
            return None
        scalar_indices = isinstance(indices, numbers.Integral)
        axis = self.axis.val
        if scalar_indices:
            res = np.take(x, [indices], axis)
            res2 = np.squeeze(res, axis=axis)
            if isinstance(res2, np.ndarray) and len(res2.shape) == 0:
                # res2 is a scalar, but represented as np.array(symbol,
                # dtype=np.object) which np.squeeze can't remove.
                return res2.item()
            return res2
        return np.take(x, indices, axis)

    def type_inference(self):
        out_type = self.x.dtype

        if self.axis.val < -self.x.rank or self.axis.val >= self.x.rank:
            raise IndexError(
                "Axis value {} is out of bounds for {} node {}".format(
                    self.axis.val, self.op_type, self.name
                )
            )

        output_rank = self.x.rank - 1 + self.indices.rank
        if output_rank == 0:
            # output scalar
            return out_type

        axis = self.axis.val
        axis = axis if axis >= 0 else axis + self.x.rank
        out_shape = self.x.shape[:axis] + self.indices.shape + self.x.shape[axis + 1 :]
        return types.tensor(out_type, out_shape)


@register_op(doc_str="")
class scatter(Operation):
    """
    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis``
    by operation ``mode``.
    
    Example: ``mode == update``.
    
    * For ``i`` in ``[0, len(indices)]``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
    .. math::
       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
    
    * For ``j! = i``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
    .. math::
       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
    
    Example: ``mode == add``.
    
    * For ``i`` in ``[0, len(indices)]``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
    .. math::
       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
    .. math::
       x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
    
    * For ``j! = i``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
    .. math::
       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
    
    Parameters
    ----------
    data: tensor<\*D, T> (Required)
    indices: tensor<[C],T> (Required)
        * 1-D tensor.
    updates: tensor<\*K, T> (Required)
        * ``K = data.shape[:axis] + [len(indices)] + data.shape[axis+1:]``.
    axis: const i32 (Optional)
        * Default to ``0``.
    mode: const string (Optional)
        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
          ``div``, ``max``, ``min``.
    
    Returns
    -------
    tensor<\*D, T>
        * With the same type and shape as input ``x``.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        data=TensorInputType(),
        indices=IntTensorInputType(),
        updates=TensorInputType(),
        axis=IntInputType(const=True, optional=True),
        mode=StringInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axis=0,
            mode="add",
            )

    def __init__(self, **kwargs):
        super(scatter, self).__init__(**kwargs)

    def type_inference(self):
        if self.axis.val < -self.data.rank or self.axis.val >= self.data.rank:
            raise IndexError(
                "Axis value {} is out of bounds for {} node {}".format(
                    self.axis.val, self.op_type, self.name
                )
            )

        axis = self.axis.val
        axis = axis if axis >= 0 else axis + self.data.rank
        expected_updates_shape = (
            self.data.shape[:axis] + self.indices.shape + self.data.shape[axis + 1 :]
        )
        np.testing.assert_equal(self.updates.shape, np.array(expected_updates_shape))

        return self.data.sym_type


@register_op(doc_str="")
class gather_along_axis(Operation):
    """
    Take the values along ``axis`` at locations ``indices``.
    
    .. math::
       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
    .. math::
       output[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] = = x[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D]
    
    Parameters
    ----------
    x: tensor<\*D, T> (Required)
    indices: tensor<\*K, T> (Required)
        * ``rank(indices) == rank(x)``.
    axis: const i32 (Optional):
        * Default to ``0``.
    
    Returns
    -------
    tensor<\*D, T>:
        * Output tensor has the same shape as ``indices``.
    
    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        indices=IntTensorInputType(),
        axis=IntInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axis=0,
            )

    def __init__(self, **kwargs):
        super(gather_along_axis, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        x = self.x.val
        indices = self.indices.val
        axis = self.axis.val
        return np.take_along_axis(x, indices, axis)

    def type_inference(self):

        if self.x.rank != self.indices.rank:
            raise ValueError(
                "Rank mismatch between input and indices. \
                              Input rank: {}, indices rank: {}".format(
                    self.x.rank, self.indices.rank
                )
            )

        if self.axis.val < -self.x.rank or self.axis.val >= self.x.rank:
            raise IndexError(
                "Axis value {} is out of bounds for {} node {}".format(
                    self.axis.val, self.op_type, self.name
                )
            )

        axis = self.axis.val
        axis = axis if axis >= 0 else axis + self.x.rank

        for i in range(self.x.rank):
            if i != axis:
                assert self.x.shape[i] == self.indices.shape[i]

        return types.tensor(self.x.dtype, self.indices.shape)


@register_op(doc_str="")
class scatter_along_axis(Operation):
    """
    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis``
    by operation ``mode``.
    
    Example: ``mode == update``.
    
    * For ``i`` in ``[0, len(indices)]``:
    
    .. math::
       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
    .. math::
       output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
    .. math::
       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
    
    * For ``j! = i``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
    .. math::
       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
    
    Example: ``mode == add``.
    
    * For ``i`` in ``[0, len(indices)]``:
    
    .. math::
       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
    .. math::
       output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
    .. math::
       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
    .. math::
       x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
    
    * For ``j! = i``:
    
    .. math::
       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
    .. math::
       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
    
    Parameters
    ----------
    data: tensor<\*D, T> (Required)
    indices: tensor<\*K,T> (Required)
        * ``rank(indices) == rank(data)``.
    updates: tensor<\*K, T> (Required)
        * Must be the same shape as ``indices``.
    axis: const i32 (Optional)
        * Default to ``0``.
    mode: const string (Optional)
        * Default to ``add``.
        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
          ``div``, ``max``, ``min``.
    
    Returns
    -------
    tensor<\*D, T>
        * With the same type and shape as input ``x``.

    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        data=TensorInputType(),
        indices=IntTensorInputType(),
        updates=TensorInputType(),
        axis=IntInputType(const=True, optional=True),
        mode=StringInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axis=0,
            mode="add",
            )

    def __init__(self, **kwargs):
        super(scatter_along_axis, self).__init__(**kwargs)

    @precondition(allow=VALUE)
    def value_inference(self):
        data = np.copy(self.data.val)
        indices = self.indices.val
        updates = self.updates.val
        axis = self.axis.val
        np_output = data
        np.put_along_axis(np_output, indices, updates, axis=axis)
        return np_output

    def type_inference(self):
        if self.axis.val < -self.data.rank or self.axis.val >= self.data.rank:
            raise IndexError(
                "Axis value {} is out of bounds for {} node {}".format(
                    self.axis.val, self.op_type, self.name
                )
            )

        axis = self.axis.val
        axis = axis if axis >= 0 else axis + self.data.rank

        assert self.indices.shape == self.updates.shape
        assert self.data.rank == self.indices.rank
        for i in range(self.data.rank):
            if i != axis:
                assert self.data.shape[i] == self.indices.shape[i]

        return self.data.sym_type


@register_op(doc_str="")
class gather_nd(Operation):
    """
    Gather slices from ``x`` according to ``indices``, similar to `tf.gather_nd <https://www.tensorflow.org/api_docs/python/tf/gather_nd>`_.
    
    The ``indices`` is a K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a slice
    of ``x``:
    
    .. math::
       output[i_0, ..., i_{K-2}]= x[indices[i_0, ..., i_{K-2}]]
    
    Where ``K = rank(indices)`` and ``x[indices[i_0, ..., i_{K-2}]]`` has rank
    ``rank(x) - indices.shape[-1]``.
    
    Parameters
    ----------
    x: tensor<\*D,T> (Required)
    indices: tensor<\*K,i32> (Required)

    Returns
    -------
    tensor<\*V,T>
        * ``V = K[:-1] + D[K[-1]:]``, where ``D = x.shape`` and ``K = indices.shape``.

    Attributes
    ----------
    T: fp32
    
    References
    ----------
    See `tf.gather_nd <https://www.tensorflow.org/api_docs/python/tf/gather_nd>`_.
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        indices=IntTensorInputType(),
        )

    def __init__(self, **kwargs):
        super(gather_nd, self).__init__(**kwargs)

    def type_inference(self):
        assert self.indices.shape[-1] <= self.x.rank
        out_type = self.x.dtype
        out_shape = self.indices.shape[:-1] + self.x.shape[self.indices.shape[-1] :]
        return types.tensor(out_type, out_shape)


@register_op(doc_str="")
class scatter_nd(Operation):
    """
    Scatter ``updates`` to ``data`` at locations ``indices``.

    The ``indices`` is a K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a
    slice of ``data``, ``K = rank(indices)``, and ``data[indices[i_0, ..., i_{K-2}]]``
    has rank ``rank(data) - indices.shape[-1]``.
    
    * Example: ``mode == update``: The ``output`` is set to ``data`` initially, and
      the op updates ``output`` as follows:
    
    .. math::
       output[indices[i_0, ..., i_{K-2}]]= updates[indices[i_0, ..., i_{K-2}]]
    
    * Example: ``mode == add``. The update rule is:
    
    .. math::
       output[indices[i_0, ..., i_{K-2}]] += updates[indices[i_0, ..., i_{K-2}]]
    
    Parameters
    ----------
    data: tensor<\*D,T> (Required)
    indices: tensor<\*K,i32> (Required)
    updates: tensor<\*K, T> (Required)
        * Must be the shape as ``K[:-1]+data.shape[K[-1]:]``.
    mode: const string (Optional)
        * Default to ``add``.
        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
          ``div``, ``max``, ``min``.
    
    Returns
    -------
    tensor<\*D,T>
        * A tensor with the same shape and type as ``data``.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        data=TensorInputType(),
        indices=IntTensorInputType(),
        updates=TensorInputType(),
        mode=StringInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            mode="add",
            )

    def __init__(self, **kwargs):
        super(scatter_nd, self).__init__(**kwargs)

    def type_inference(self):
        assert self.indices.shape[-1] <= self.data.rank
        expected_updates_shape = (
            self.indices.shape[:-1] + self.data.shape[self.indices.shape[-1] :]
        )
        assert is_compatible_symbolic_vector(
            self.updates.shape, tuple(expected_updates_shape)
        )
        return self.data.sym_type



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

"""
Utilities to annotate Neural Network Features with flexible shape information.
Only available in coremltools 2.0b1 and onwards
"""

from ..utils import _get_feature, _get_nn_layers, _get_input_names
from ... import _MINIMUM_FLEXIBLE_SHAPES_SPEC_VERSION
from ... import _MINIMUM_NDARRAY_SPEC_VERSION

_SEQUENCE_KEY = "S"
_BATCH_KEY = "B"
_CHANNEL_KEY = "C"
_HEIGHT_KEY = "H"
_WIDTH_KEY = "W"

_CONSTRAINED_KEYS = [_CHANNEL_KEY, _HEIGHT_KEY, _WIDTH_KEY]


class Shape(object):
    def __init__(self, shape_value):
        if shape_value < 1:
            raise Exception("Invalid value. Size/Shape values must be > 0")
        self._value = shape_value

    @property
    def value(self):
        return self._value


class Size(Shape):
    def __init__(self, size_value):
        super(Size, self).__init__(size_value)


class NeuralNetworkMultiArrayShape:
    """
    An object representing a shape for a multiArray feature in a
    neural network. Valid shapes must have have only the Channel [C]
    shape or the Channel, Height and Width [C, H, W] shapes populated
    """

    def __init__(self, channel=None, height=None, width=None):
        self._shape = {
            _CHANNEL_KEY: Shape(int(channel)) if channel else None,
            _HEIGHT_KEY: Shape(int(height)) if height else None,
            _WIDTH_KEY: Shape(int(width)) if width else None,
        }

    def set_channel_shape(self, channel_shape):
        self._shape[_CHANNEL_KEY] = Shape(channel_shape)

    def set_height_shape(self, height_shape):
        self._shape[_HEIGHT_KEY] = Shape(height_shape)

    def set_width_shape(self, width_shape):
        self._shape[_WIDTH_KEY] = Shape(width_shape)

    def _validate_multiarray_shape(self):
        num_dims = len([v for v in self._shape.values() if v])
        if num_dims != 1 and num_dims != 3:
            raise Exception(
                "For neural networks, shape must be of length 1 or 3"
                ", representing input shape [C] or [C,H,W], respectively"
            )

        if num_dims == 1:
            if not self._shape["C"]:
                raise Exception("Channel Shape not specified")

    @property
    def multiarray_shape(self):
        num_dims = len([v for v in self._shape.values() if v])
        if num_dims == 1:
            return [self._shape[_CHANNEL_KEY].value]
        elif num_dims == 3:
            return [
                self._shape[_CHANNEL_KEY].value,
                self._shape[_HEIGHT_KEY].value,
                self._shape[_WIDTH_KEY].value,
            ]
        else:
            raise Exception("Invalid multiarray shape for neural network")


class NeuralNetworkImageSize:
    """
    An object representing a size for an image feature inside a
    neural network. Valid sizess for height and width are > 0.
    """

    def __init__(self, height=None, width=None):
        self._height = Size(height)
        self._width = Size(width)

    def set_width(self, width):
        self._width = Size(width)

    def set_height(self, height):
        self._height = Size(height)

    @property
    def width(self):
        return self._width.value

    @property
    def height(self):
        return self._height.value


class ShapeRange(object):
    def __init__(self, lowerBound, upperBound):
        unBounded = False

        if upperBound == -1:
            unBounded = True

        if not unBounded and lowerBound > upperBound:
            raise Exception(
                "lowerBound > upperBound for range ({},{})".format(
                    lowerBound, upperBound
                )
            )

        if not unBounded and upperBound < 1:
            raise Exception("Invalid upperBound: {} ".format(upperBound))

        if lowerBound == 0:
            lowerBound = 1

        if lowerBound < 1:
            raise Exception("Invalid lowerBound: {}".format(lowerBound))

        self._lowerBound = lowerBound
        self._upperBound = upperBound
        self._unBounded = unBounded

    @property
    def lowerBound(self):
        return self._lowerBound

    @property
    def upperBound(self):
        return self._upperBound

    @property
    def isUnbounded(self):
        return self._unBounded

    @property
    def isFlexible(self):
        return not (self._lowerBound == self._upperBound)


class NeuralNetworkMultiArrayShapeRange:
    """
    An object representing a range of shapes for a multiArray feature in a
    neural network. Valid shape ranges must have have only the Channel [C]
    range or the Channel, Height and Width [C, H, W] ranges populated. A "-1"
    value in an upper bound represents an unbounded range.
    """

    def __init__(self, input_ranges=None):
        self.arrayShapeRange = {}

        if input_ranges:
            if not isinstance(input_ranges, dict):
                raise Exception(
                    "Attempting to initialize a shape range with something other than a dictionary of shapes."
                )
            self.arrayShapeRange = {}
            for key, value in input_ranges.items():
                if key in _CONSTRAINED_KEYS:
                    self.arrayShapeRange[key] = self._create_shape_range(value)
            self.validate_array_shape_range()

    def _create_shape_range(self, r):
        if not isinstance(r, tuple):
            raise Exception("Range should be a ShapeRange or a tuple object")
        elif len(r) != 2:
            raise Exception("Range tuple should be at least length 2")
        return ShapeRange(r[0], r[1])

    def add_channel_range(self, channel_range):
        if not isinstance(channel_range, ShapeRange):
            channel_range = self._create_shape_range(channel_range)
        self.arrayShapeRange[_CHANNEL_KEY] = channel_range

    def add_height_range(self, height_range):
        if not isinstance(height_range, ShapeRange):
            height_range = self._create_shape_range(height_range)
        self.arrayShapeRange[_HEIGHT_KEY] = height_range

    def add_width_range(self, width_range):
        if not isinstance(width_range, ShapeRange):
            width_range = self._create_shape_range(width_range)
        self.arrayShapeRange[_WIDTH_KEY] = width_range

    def get_shape_range_dims(self):
        return len(self.arrayShapeRange.keys())

    def validate_array_shape_range(self):
        num_dims = self.get_shape_range_dims()
        if num_dims != 1 and num_dims != 3:
            raise Exception(
                "For neural networks, shape must be of length 1 or 3"
                ", representing input shape [C] or [C,H,W], respectively"
            )

        if num_dims == 1:
            if _CHANNEL_KEY not in self.arrayShapeRange.keys():
                raise Exception("Channel Shape Range not specified")

        if num_dims == 3:
            if (
                _CHANNEL_KEY not in self.arrayShapeRange.keys()
                or _HEIGHT_KEY not in self.arrayShapeRange.keys()
                or _WIDTH_KEY not in self.arrayShapeRange.keys()
            ):
                raise Exception(
                    "Shape range constraint missing for either channel, height, or width."
                )

    def get_channel_range(self):
        return self.arrayShapeRange[_CHANNEL_KEY]

    def get_height_range(self):
        return self.arrayShapeRange[_HEIGHT_KEY]

    def get_width_range(self):
        return self.arrayShapeRange[_WIDTH_KEY]

    def isFlexible(self):
        """
        Returns true if any one of the channel, height, or width ranges of this shape allow more than one input value.
        """
        for key, value in self.arrayShapeRange.items():
            if key in _CONSTRAINED_KEYS:
                if value.isFlexible:
                    return True

        return False


class NeuralNetworkImageSizeRange:
    """
    An object representing a range of sizes for an image feature inside a
    neural network. Valid ranges for height and width are > 0. A "-1"
    upper bound value for either width or height represents an unbounded size
    for that dimension.
    """

    def __init__(self, height_range=None, width_range=None):
        if height_range and not isinstance(height_range, ShapeRange):
            if not isinstance(height_range, tuple):
                raise Exception("Height range should be a ShapeRange or a tuple object")
            elif len(height_range) != 2:
                raise Exception("Height range tuple should be at least length 2")
            height_range = ShapeRange(height_range[0], height_range[1])

        if width_range and not isinstance(width_range, ShapeRange):
            if not isinstance(width_range, tuple):
                raise Exception("Width range should be a ShapeRange or a tuple object")
            elif len(width_range) != 2:
                raise Exception("Width range tuple should be at least length 2")
            width_range = ShapeRange(width_range[0], width_range[1])

        self._height_range = height_range
        self._width_range = width_range

    def add_width_range(self, width_range):
        if not isinstance(width_range, ShapeRange):
            if not isinstance(width_range, tuple):
                raise Exception("Width range should be a ShapeRange or a tuple object")
            elif len(width_range) != 2:
                raise Exception("Width range tuple should be at least length 2")

        self._width_range = ShapeRange(width_range[0], width_range[1])

    def add_height_range(self, height_range):
        if not isinstance(height_range, ShapeRange):
            if not isinstance(height_range, tuple):
                raise Exception("Height range should be a ShapeRange or a tuple object")
            elif len(height_range) != 2:
                raise Exception("Height range tuple should be at least length 2")

        self._height_range = ShapeRange(height_range[0], height_range[1])

    def get_width_range(self):
        return self._width_range

    def get_height_range(self):
        return self._height_range


def add_enumerated_multiarray_shapes(spec, feature_name, shapes):
    """
    Annotate an input or output multiArray feature in a Neural Network spec to
    to accommodate a list of enumerated array shapes

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the image feature for which to add shape information.
        If the feature is not found in the input or output descriptions then
        an exception is thrown

    :param shapes: [] | NeuralNetworkMultiArrayShape
        A single or a list of NeuralNetworkImageSize objects which encode valid
        size information for a image feature

    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> array_shapes = [flexible_shape_utils.NeuralNetworkMultiArrayShape(3)]
        >>> second_shape = flexible_shape_utils.NeuralNetworkMultiArrayShape()
        >>> second_shape.set_channel_shape(3)
        >>> second_shape.set_height_shape(10)
        >>> second_shape.set_width_shape(15)
        >>> array_shapes.append(second_shape)
        >>> flexible_shape_utils.add_enumerated_multiarray_shapes(spec, feature_name='my_multiarray_featurename', shapes=array_shapes)

    :return:
        None. The spec object is updated
    """

    if not isinstance(shapes, list):
        shapes = [shapes]

    for shape in shapes:
        if not isinstance(shape, NeuralNetworkMultiArrayShape):
            raise Exception(
                "Shape ranges should be of type NeuralNetworkMultiArrayShape"
            )
        shape._validate_multiarray_shape()

    feature = _get_feature(spec, feature_name)
    if feature.type.WhichOneof("Type") != "multiArrayType":
        raise Exception(
            "Trying to add enumerated shapes to " "a non-multiArray feature type"
        )

    if feature.type.multiArrayType.WhichOneof("ShapeFlexibility") != "enumeratedShapes":
        feature.type.multiArrayType.ClearField("ShapeFlexibility")

    eshape_len = len(feature.type.multiArrayType.enumeratedShapes.shapes)

    # Add default array shape to list of enumerated shapes if enumerated shapes
    # field is currently empty
    if eshape_len == 0:
        fixed_shape = feature.type.multiArrayType.shape
        if len(fixed_shape) == 1:
            fs = NeuralNetworkMultiArrayShape(fixed_shape[0])
            shapes.append(fs)
        elif len(fixed_shape) == 3:
            fs = NeuralNetworkMultiArrayShape()
            fs.set_channel_shape(fixed_shape[0])
            fs.set_height_shape(fixed_shape[1])
            fs.set_width_shape(fixed_shape[2])
            shapes.append(fs)
        else:
            raise Exception(
                "Original fixed multiArray shape for {} is invalid".format(feature_name)
            )

    for shape in shapes:
        s = feature.type.multiArrayType.enumeratedShapes.shapes.add()
        s.shape.extend(shape.multiarray_shape)

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_FLEXIBLE_SHAPES_SPEC_VERSION, spec.specificationVersion
    )


def add_enumerated_image_sizes(spec, feature_name, sizes):
    """
    Annotate an input or output image feature in a Neural Network spec to
    to accommodate a list of enumerated image sizes

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the image feature for which to add size information.
        If the feature is not found in the input or output descriptions then
        an exception is thrown

    :param sizes:  [] | NeuralNetworkImageSize
        A single or a list of NeuralNetworkImageSize objects which encode valid
        size information for a image feature

    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> image_sizes = [flexible_shape_utils.NeuralNetworkImageSize(128, 128)]
        >>> image_sizes.append(flexible_shape_utils.NeuralNetworkImageSize(256, 256))
        >>> flexible_shape_utils.add_enumerated_image_sizes(spec, feature_name='my_multiarray_featurename', sizes=image_sizes)

    :return:
        None. The spec object is updated
    """
    if not isinstance(sizes, list):
        sizes = [sizes]

    for size in sizes:
        if not isinstance(size, NeuralNetworkImageSize):
            raise Exception("Shape ranges should be of type NeuralNetworkImageSize")

    feature = _get_feature(spec, feature_name)
    if feature.type.WhichOneof("Type") != "imageType":
        raise Exception("Trying to add enumerated sizes to " "a non-image feature type")

    if feature.type.imageType.WhichOneof("SizeFlexibility") != "enumeratedSizes":
        feature.type.imageType.ClearField("SizeFlexibility")

    esizes_len = len(feature.type.imageType.enumeratedSizes.sizes)

    # Add default image size to list of enumerated sizes if enumerated sizes
    # field is currently empty
    if esizes_len == 0:
        fixed_height = feature.type.imageType.height
        fixed_width = feature.type.imageType.width
        sizes.append(NeuralNetworkImageSize(fixed_height, fixed_width))

    for size in sizes:
        s = feature.type.imageType.enumeratedSizes.sizes.add()
        s.height = size.height
        s.width = size.width

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_FLEXIBLE_SHAPES_SPEC_VERSION, spec.specificationVersion
    )


def update_image_size_range(spec, feature_name, size_range):
    """
    Annotate an input or output Image feature in a Neural Network spec to
    to accommodate a range of image sizes

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the Image feature for which to add shape information.
        If the feature is not found in the input or output descriptions then
        an exception is thrown

    :param size_range: NeuralNetworkImageSizeRange
        A NeuralNetworkImageSizeRange object with the populated image size
        range information.

    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> img_size_ranges = flexible_shape_utils.NeuralNetworkImageSizeRange()
        >>> img_size_ranges.add_height_range(64, 128)
        >>> img_size_ranges.add_width_range(128, -1)
        >>> flexible_shape_utils.update_image_size_range(spec, feature_name='my_multiarray_featurename', size_range=img_size_ranges)

    :return:
        None. The spec object is updated
    """
    if not isinstance(size_range, NeuralNetworkImageSizeRange):
        raise Exception("Shape ranges should be of type NeuralNetworkImageSizeRange")

    feature = _get_feature(spec, feature_name)
    if feature.type.WhichOneof("Type") != "imageType":
        raise Exception("Trying to add size ranges for " "a non-image feature type")

    feature.type.imageType.ClearField("SizeFlexibility")
    feature.type.imageType.imageSizeRange.heightRange.lowerBound = (
        size_range.get_height_range().lowerBound
    )
    feature.type.imageType.imageSizeRange.heightRange.upperBound = (
        size_range.get_height_range().upperBound
    )

    feature.type.imageType.imageSizeRange.widthRange.lowerBound = (
        size_range.get_width_range().lowerBound
    )
    feature.type.imageType.imageSizeRange.widthRange.upperBound = (
        size_range.get_width_range().upperBound
    )

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_FLEXIBLE_SHAPES_SPEC_VERSION, spec.specificationVersion
    )


def update_multiarray_shape_range(spec, feature_name, shape_range):
    """
    Annotate an input or output MLMultiArray feature in a Neural Network spec
    to accommodate a range of shapes

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the feature for which to add shape range
        information. If the feature is not found in the input or output
        descriptions then an exception is thrown

    :param shape_range: NeuralNetworkMultiArrayShapeRange
        A NeuralNetworkMultiArrayShapeRange object with the populated shape
        range information. The shape_range object must either contain only
        shape information for channel or channel, height and width. If
        the object is invalid then an exception is thrown

    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> shape_range = flexible_shape_utils.NeuralNetworkMultiArrayShapeRange()
        >>> shape_range.add_channel_range((1, 3))
        >>> shape_range.add_width_range((128, 256))
        >>> shape_range.add_height_range((128, 256))
        >>> flexible_shape_utils.update_multiarray_shape_range(spec, feature_name='my_multiarray_featurename', shape_range=shape_range)

    :return:
        None. The spec is updated
    """
    if not isinstance(shape_range, NeuralNetworkMultiArrayShapeRange):
        raise Exception("Shape range should be of type MultiArrayShapeRange")

    shape_range.validate_array_shape_range()
    feature = _get_feature(spec, feature_name)

    if feature.type.WhichOneof("Type") != "multiArrayType":
        raise Exception(
            "Trying to update shape range for " "a non-multiArray feature type"
        )

    # Add channel range
    feature.type.multiArrayType.ClearField("ShapeFlexibility")
    s = feature.type.multiArrayType.shapeRange.sizeRanges.add()
    s.lowerBound = shape_range.get_channel_range().lowerBound
    s.upperBound = shape_range.get_channel_range().upperBound

    if shape_range.get_shape_range_dims() > 1:
        # Add height range
        s = feature.type.multiArrayType.shapeRange.sizeRanges.add()
        s.lowerBound = shape_range.get_height_range().lowerBound
        s.upperBound = shape_range.get_height_range().upperBound
        # Add width range
        s = feature.type.multiArrayType.shapeRange.sizeRanges.add()
        s.lowerBound = shape_range.get_width_range().lowerBound
        s.upperBound = shape_range.get_width_range().upperBound

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_FLEXIBLE_SHAPES_SPEC_VERSION, spec.specificationVersion
    )


def set_multiarray_ndshape_range(spec, feature_name, lower_bounds, upper_bounds):
    """
    Annotate an input or output MLMultiArray feature in a Neural Network spec
    to accommodate a range of shapes.
    This is different from "update_multiarray_shape_range", which works with rank 5
    SBCHW mapping.

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the feature for which to add shape range
        information. If the feature is not found in the input or output
        descriptions then an exception is thrown

    :param lower_bounds: List[int]
        list of integers specifying the lower bounds of each dimension.
        Length must be same as the rank (length of shape) of the feature_name.

    :param upper_bounds: List[int]
        list of integers specifying the upper bounds of each dimension.
        -1 corresponds to unbounded range.
        Length must be same as the rank (length of shape) of the feature_name.


    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> # say, the default shape of "my_multiarray_featurename" is (2,3)
        >>> flexible_shape_utils.set_multiarray_ndshape_range(spec, feature_name='my_multiarray_featurename', lower_bounds=[1,2], upper_bounds=[10,-1])

    :return:
        None. The spec is updated
    """
    if not isinstance(lower_bounds, list):
        raise Exception("lower_bounds must be a list")
    if not isinstance(upper_bounds, list):
        raise Exception("upper_bounds must be a list")

    feature = _get_feature(spec, feature_name)

    if feature.type.WhichOneof("Type") != "multiArrayType":
        raise Exception(
            "Trying to update shape range for " "a non-multiArray feature type"
        )

    shape = feature.type.multiArrayType.shape

    if len(shape) != len(lower_bounds):
        raise Exception(
            "Length of lower_bounds is not equal to the number of dimensions in the default shape"
        )
    if len(shape) != len(upper_bounds):
        raise Exception(
            "Length of upper_bounds is not equal to the number of dimensions in the default shape"
        )

    feature.type.multiArrayType.ClearField("ShapeFlexibility")

    for i in range(len(lower_bounds)):
        if shape[i] < lower_bounds[i]:
            raise Exception(
                "Default shape in %d-th dimension, which is %d, is smaller"
                " than the lower bound of %d" % (i, int(shape[i]), lower_bounds[i])
            )
        if upper_bounds[i] != -1:
            if shape[i] > upper_bounds[i]:
                raise Exception(
                    "Default shape in %d-th dimension, which is %d, is greater"
                    " than the upper bound of %d" % (i, int(shape[i]), upper_bounds[i])
                )

        s = feature.type.multiArrayType.shapeRange.sizeRanges.add()
        s.lowerBound = lower_bounds[i]
        s.upperBound = upper_bounds[i]

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_NDARRAY_SPEC_VERSION, spec.specificationVersion
    )


def add_multiarray_ndshape_enumeration(spec, feature_name, enumerated_shapes):
    """
    Annotate an input or output MLMultiArray feature in a Neural Network spec
    to accommodate a range of shapes.
    Add provided enumerated shapes to the list of shapes already present.
    This method is different from "add_enumerated_multiarray_shapes", which is applicable
    for rank 5 mapping, SBCHW, arrays.

    :param spec: MLModel
        The MLModel spec containing the feature

    :param feature_name: str
        The name of the feature for which to add shape range
        information. If the feature is not found in the input or output
        descriptions then an exception is thrown

    :param enumerated_shapes: List[Tuple(int)]
        list of shapes, where each shape is specified as a tuple of integers.


    Examples
    --------
    .. sourcecode:: python

        >>> import coremltools
        >>> from coremltools.models.neural_network import flexible_shape_utils
        >>> spec = coremltools.utils.load_spec('mymodel.mlmodel')
        >>> # say, the default shape of "my_multiarray_featurename" is (2,3)
        >>> flexible_shape_utils.add_multiarray_ndshape_enumeration(spec, feature_name='my_multiarray_featurename', enumerated_shapes=[(2,4), (2,6)])

    :return:
        None. The spec is updated
    """
    if not isinstance(enumerated_shapes, list):
        raise Exception("enumerated_shapes must be a list")
    if len(enumerated_shapes) == 0:
        raise Exception("enumerated_shapes is empty")

    feature = _get_feature(spec, feature_name)
    if feature.type.WhichOneof("Type") != "multiArrayType":
        raise Exception(
            "Trying to update shape range for " "a non-multiArray feature type"
        )

    shape = feature.type.multiArrayType.shape

    if feature.type.multiArrayType.WhichOneof("ShapeFlexibility") != "enumeratedShapes":
        feature.type.multiArrayType.ClearField("ShapeFlexibility")

    eshape_len = len(feature.type.multiArrayType.enumeratedShapes.shapes)

    # Add default array shape to list of enumerated shapes if enumerated shapes
    # field is currently empty
    if eshape_len == 0:
        fixed_shape = feature.type.multiArrayType.shape
        s = feature.type.multiArrayType.enumeratedShapes.shapes.add()
        s.shape.extend(fixed_shape)

    for shape in enumerated_shapes:
        if not isinstance(shape, tuple):
            raise Exception("An element in 'enumerated_shapes' is not a tuple")
        s = feature.type.multiArrayType.enumeratedShapes.shapes.add()
        s.shape.extend(list(shape))

    # Bump up specification version
    spec.specificationVersion = max(
        _MINIMUM_NDARRAY_SPEC_VERSION, spec.specificationVersion
    )



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from . import _sklearn_util

from ... import SPECIFICATION_VERSION
from ...models._interface_management import set_transform_interface_params

from ...proto import Model_pb2 as _Model_pb2
from ...proto import FeatureTypes_pb2 as _FeatureTypes_pb2
from ...proto import OneHotEncoder_pb2 as _OHE_pb2
from ...models import datatypes
from ...models import MLModel as _MLModel
from ...models.feature_vectorizer import create_feature_vectorizer
from ...models.array_feature_extractor import create_array_feature_extractor

from ...models.pipeline import Pipeline

from ..._deps import _HAS_SKLEARN as _HAS_SKLEARN

if _HAS_SKLEARN:
    import sklearn
    from sklearn.preprocessing import OneHotEncoder

    sklearn_class = OneHotEncoder

# model type determines the behavior of this module.
model_type = "transformer"


def convert(model, input_features, output_features):
    """Convert a one-hot-encoder model to the protobuf spec.

    Parameters
    ----------
    model: OneHotEncoder
        A trained one-hot encoder model.

    input_features: str, optional
        Name of the input column.

    output_features: str, optional
        Name of the output column.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    # Make sure the model is fitted.
    _sklearn_util.check_expected_type(model, OneHotEncoder)
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "active_features_"))
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "n_values_"))

    input_dimension = get_input_dimension(model)

    if input_dimension is not None:
        # Make sure that our starting dimensions are correctly managed.
        assert len(input_features) == 1
        assert input_features[0][1] == datatypes.Array(input_dimension)

    input_dimension = input_features[0][1].num_elements

    expected_output_dimension = update_dimension(model, input_dimension)
    assert output_features[0][1] == datatypes.Array(expected_output_dimension)

    # Create a pipeline that can do all of the subsequent feature extraction.
    feature_vectorizer_input_features = []
    feature_vectorizer_size_map = {}

    if model.categorical_features == "all":
        _categorical_features = set(range(input_dimension))
        _cat_feature_idx_mapping = dict((i, i) for i in range(input_dimension))
    else:
        _categorical_features = set(model.categorical_features)
        _cat_feature_idx_mapping = dict(
            (_idx, i) for i, _idx in enumerate(sorted(model.categorical_features))
        )

    pline = Pipeline(input_features, output_features)

    # Track the overall packing index, which determines the output ordering.
    pack_idx = 0

    # First, go through all the columns that are encoded. The sklearn OHE puts
    # all of these first, regardless of their original ordering.
    for idx in range(input_dimension):
        f_name = "__OHE_%d__" % pack_idx

        if idx in _categorical_features:

            # This input column is one hot encoded
            feature_extractor_spec = create_array_feature_extractor(
                input_features, f_name, idx, output_type="Int64"
            )

            pline.add_model(feature_extractor_spec)

            _cat_feature_idx = _cat_feature_idx_mapping[idx]

            ohe_input_features = [(f_name, datatypes.Int64())]
            ohe_output_features = [(f_name, datatypes.Dictionary("Int64"))]

            # Create a one hot encoder per column
            o_spec = _Model_pb2.Model()
            o_spec.specificationVersion = SPECIFICATION_VERSION
            o_spec = set_transform_interface_params(
                o_spec, ohe_input_features, ohe_output_features
            )

            ohe_spec = o_spec.oneHotEncoder
            ohe_spec.outputSparse = True

            if model.handle_unknown == "error":
                ohe_spec.handleUnknown = _OHE_pb2.OneHotEncoder.HandleUnknown.Value(
                    "ErrorOnUnknown"
                )
            else:
                ohe_spec.handleUnknown = _OHE_pb2.OneHotEncoder.HandleUnknown.Value(
                    "IgnoreUnknown"
                )

            # Need to do a quick search to find the part of the active_features_ mask
            # that represents the categorical variables in our part.  Could do this
            # with binary search, but we probably don't need speed so much here.
            def bs_find(a, i):
                lb, k = 0, len(a)
                while k > 0:
                    _idx = lb + (k // 2)
                    if a[_idx] < i:
                        lb = _idx + 1
                        k -= 1
                    k = k // 2

                return lb

            # Here are the indices we are looking fo
            f_idx_bottom = model.feature_indices_[_cat_feature_idx]
            f_idx_top = model.feature_indices_[_cat_feature_idx + 1]

            # Now find where in the active features list we should look.
            cat_feat_idx_bottom = bs_find(model.active_features_, f_idx_bottom)
            cat_feat_idx_top = bs_find(model.active_features_, f_idx_top)
            n_cat_values = cat_feat_idx_top - cat_feat_idx_bottom

            for i in range(cat_feat_idx_bottom, cat_feat_idx_top):
                # The actual categorical value is stored as an offset in the active_features list.
                cat_idx = model.active_features_[i] - f_idx_bottom
                ohe_spec.int64Categories.vector.append(cat_idx)

            # Add the ohe to the pipeline
            pline.add_model(o_spec)

            # Add the result to the feature_vectorizer at the end.
            feature_vectorizer_input_features.append(
                (f_name, datatypes.Dictionary("Int64"))
            )
            feature_vectorizer_size_map[f_name] = n_cat_values

            pack_idx += 1

    # Now go through all the columns that are not encoded as the sklearn OHE puts
    # these after the encoded ones.  For speed, we can put these all in a single
    # ArrayFeatureExtractor
    #
    pass_through_features = [
        idx for idx in range(input_dimension) if idx not in _categorical_features
    ]

    if pass_through_features:
        f_name = "__OHE_pass_through__"

        # This input column is not one hot encoded
        feature_extractor_spec = create_array_feature_extractor(
            input_features, f_name, pass_through_features
        )

        pline.add_model(feature_extractor_spec)
        feature_vectorizer_input_features.append(
            (f_name, datatypes.Array(len(pass_through_features)))
        )

    # Finally, add the feature vectorizer to the pipeline.
    output_feature_name = output_features[0][0]
    output_feature_dimension = output_features[0][1].num_elements

    fvec, _num_out_dim = create_feature_vectorizer(
        feature_vectorizer_input_features,
        output_features[0][0],
        feature_vectorizer_size_map,
    )

    # Make sure that the feature vectorizer input actually matches up with the
    assert _num_out_dim == output_features[0][1].num_elements

    pline.add_model(fvec)

    return _MLModel(pline.spec)


def update_dimension(model, input_dimension):
    """
    Given a model that takes an array of dimension input_dimension, returns
    the output dimension.
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "active_features_"))
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "n_values_"))

    if model.categorical_features == "all":
        return len(model.active_features_)
    else:
        out_dimension = len(model.active_features_) + (
            input_dimension - len(model.n_values_)
        )

    return out_dimension


def get_input_dimension(model):
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "active_features_"))
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "n_values_"))

    if model.categorical_features == "all":
        return len(model.feature_indices_) - 1
    else:
        # This can't actually be determined from the model as indices after the
        # rest of the categorical values don't seem to be tracked
        return None



# Copyright (c) 2019, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import os
import shutil
from coremltools.models.utils import _is_macos

import unittest

from coremltools.models import MLModel
from coremltools.models.nearest_neighbors import KNearestNeighborsClassifierBuilder
from coremltools._deps import _HAS_SKLEARN

if _HAS_SKLEARN:
    from sklearn.datasets import load_iris


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class NearestNeighborsBuilderTest(unittest.TestCase):
    """
    Unit tests for the nearest neighbors builder class.
    """

    def setUp(self):
        iris_samples = load_iris()
        self.iris_X = iris_samples.data
        self.iris_y = iris_samples.target
        self.training_X = self.iris_X[-30:]
        self.training_y = self.iris_y[-30:]

    def tearDown(self):
        # Do any cleanup here
        pass

    def create_builder(self, default_class_label="default_label"):
        builder = KNearestNeighborsClassifierBuilder(
            input_name="input",
            output_name="output",
            number_of_dimensions=4,
            default_class_label=default_class_label,
        )
        return builder

    def test_builder_output_types(self):
        builder = self.create_builder(default_class_label="default")
        self.assertIsNotNone(builder)
        self.assertTrue(
            builder.spec.kNearestNeighborsClassifier.HasField("stringClassLabels")
        )

        builder = self.create_builder(default_class_label=12)
        self.assertIsNotNone(builder)
        self.assertTrue(
            builder.spec.kNearestNeighborsClassifier.HasField("int64ClassLabels")
        )

        with self.assertRaises(TypeError):
            bad_default_label = float(21.32)
            self.create_builder(default_class_label=bad_default_label)

    def test_builder_training_input(self):
        builder = self.create_builder(default_class_label="default")
        self.assertIsNotNone(builder)
        self.assertTrue(
            builder.spec.kNearestNeighborsClassifier.HasField("stringClassLabels")
        )

        self.assertEqual(builder.spec.description.trainingInput[0].name, "input")
        self.assertEqual(
            builder.spec.description.trainingInput[0].type.WhichOneof("Type"),
            "multiArrayType",
        )
        self.assertEqual(builder.spec.description.trainingInput[1].name, "output")
        self.assertEqual(
            builder.spec.description.trainingInput[1].type.WhichOneof("Type"),
            "stringType",
        )

    def test_make_updatable(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        self.assertTrue(builder.spec.isUpdatable)
        builder.is_updatable = False
        self.assertFalse(builder.spec.isUpdatable)
        builder.is_updatable = True
        self.assertTrue(builder.spec.isUpdatable)

    def test_author(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        self.assertEqual(builder.spec.description.metadata.author, "")
        builder.author = "John Doe"
        self.assertEqual(builder.author, "John Doe")
        self.assertEqual(builder.spec.description.metadata.author, "John Doe")

    def test_description(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        self.assertEqual(builder.spec.description.metadata.shortDescription, "")
        builder.description = "This is a description"
        self.assertEqual(builder.description, "This is a description")
        self.assertEqual(
            builder.spec.description.metadata.shortDescription, "This is a description"
        )

    def test_weighting_scheme(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        builder.weighting_scheme = "uniform"
        self.assertEqual(builder.weighting_scheme, "uniform")

        builder.weighting_scheme = "inverse_distance"
        self.assertEqual(builder.weighting_scheme, "inverse_distance")

        builder.weighting_scheme = "unIfOrM"
        self.assertEqual(builder.weighting_scheme, "uniform")

        builder.weighting_scheme = "InVerSE_DISTance"
        self.assertEqual(builder.weighting_scheme, "inverse_distance")

        with self.assertRaises(TypeError):
            builder.weighting_scheme = "test"

    def test_index_type(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        self.assertEqual(builder.index_type, "linear")
        self.assertEqual(builder.leaf_size, 0)

        builder.set_index_type("kd_tree")
        self.assertEqual(builder.index_type, "kd_tree")  # test default value
        self.assertEqual(builder.leaf_size, 30)

        builder.set_index_type("linear")
        self.assertEqual(builder.index_type, "linear")
        self.assertEqual(builder.leaf_size, 0)

        builder.set_index_type("kd_tree", leaf_size=45)  # test user-defined value
        self.assertEqual(builder.index_type, "kd_tree")
        self.assertEqual(builder.leaf_size, 45)

        builder.set_index_type("linear", leaf_size=37)
        self.assertEqual(builder.index_type, "linear")
        self.assertEqual(builder.leaf_size, 0)

        builder.set_index_type("KD_TrEe", leaf_size=22)  # test user-defined value
        self.assertEqual(builder.index_type, "kd_tree")
        self.assertEqual(builder.leaf_size, 22)

        builder.set_index_type("linEAR")
        self.assertEqual(builder.index_type, "linear")
        self.assertEqual(builder.leaf_size, 0)

        with self.assertRaises(TypeError):
            builder.set_index_type("unsupported_index")

        with self.assertRaises(TypeError):
            builder.set_index_type("kd_tree", -10)

        with self.assertRaises(TypeError):
            builder.set_index_type("kd_tree", 0)

    def test_leaf_size(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        builder.set_index_type("kd_tree", leaf_size=45)  # test user-defined value
        self.assertEqual(builder.index_type, "kd_tree")
        self.assertEqual(builder.leaf_size, 45)

        builder.leaf_size = 12
        self.assertEqual(builder.index_type, "kd_tree")
        self.assertEqual(builder.leaf_size, 12)

    def test_set_number_of_neighbors_with_bounds(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        self.assertEqual(builder.number_of_neighbors, 5)
        (min_value, max_value) = builder.number_of_neighbors_allowed_range()
        self.assertEqual(min_value, 1)
        self.assertEqual(max_value, 1000)

        builder.set_number_of_neighbors_with_bounds(12, allowed_range=(2, 24))
        (min_value, max_value) = builder.number_of_neighbors_allowed_range()
        self.assertEqual(builder.number_of_neighbors, 12)
        self.assertEqual(min_value, 2)
        self.assertEqual(max_value, 24)
        allowed_values = builder.number_of_neighbors_allowed_set()
        self.assertIsNone(allowed_values)

        test_set = {3, 5, 7, 9}
        builder.set_number_of_neighbors_with_bounds(7, allowed_set=test_set)
        self.assertEqual(builder.number_of_neighbors, 7)
        allowed_values = builder.number_of_neighbors_allowed_set()
        self.assertIsNotNone(allowed_values)
        self.assertEqual(allowed_values, test_set)

    def test_set_number_of_neighbors_with_bounds_error_conditions(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(3)

        test_range = (3, 15)
        test_set = {1, 3, 5}
        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(
                3, allowed_range=test_range, allowed_set=test_set
            )

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(3, allowed_range=(-5, 5))

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(3, allowed_range=(5, 1))

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(
                3, allowed_range=test_range, allowed_set=test_set
            )

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(2, allowed_range=test_range)

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(5, allowed_set={5, -3, 7})

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_set=test_set)

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_set=test_set)

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(2, allowed_set=[1, 2, 3])

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_range={2, 200})

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_range=(2, 10, 20))

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_set=set())

        with self.assertRaises(TypeError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_range=[])

    def test_set_number_of_neighbors(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        builder.set_number_of_neighbors_with_bounds(12, allowed_range=(2, 24))
        self.assertEqual(builder.number_of_neighbors, 12)

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(1, allowed_range=(2, 24))
        builder.set_number_of_neighbors_with_bounds(4, allowed_range=(2, 24))
        self.assertEqual(builder.number_of_neighbors, 4)

        test_set = {3, 5, 7, 9}
        builder.set_number_of_neighbors_with_bounds(7, allowed_set=test_set)

        with self.assertRaises(ValueError):
            builder.set_number_of_neighbors_with_bounds(4, allowed_set=test_set)
        builder.set_number_of_neighbors_with_bounds(5, allowed_set=test_set)
        self.assertEqual(builder.number_of_neighbors, 5)

    def test_add_samples_invalid_data(self):
        builder = self.create_builder()
        self.assertIsNotNone(builder)

        invalid_X = [[1.0, 2.4]]
        with self.assertRaises(TypeError):
            builder.add_samples(invalid_X, self.training_y)

        with self.assertRaises(TypeError):
            builder.add_samples(self.training_X, self.training_y[:3])

        with self.assertRaises(TypeError):
            builder.add_samples([], self.training_y)

        with self.assertRaises(TypeError):
            builder.add_samples(self.training_X, [])

    def test_add_samples_int_labels(self):
        builder = self.create_builder(default_class_label=12)
        self.assertIsNotNone(builder)

        some_X = self.training_X[:10]
        some_y = self.training_y[:10]
        builder.add_samples(some_X, some_y)
        self._validate_samples(builder.spec, some_X, some_y)

        addl_X = self.training_X[10:20]
        addl_y = self.training_y[10:20]
        builder.add_samples(addl_X, addl_y)
        self._validate_samples(builder.spec, self.training_X[:20], self.training_y[:20])

    def test_add_samples_string_labels(self):
        builder = self.create_builder(default_class_label="default")
        self.assertIsNotNone(builder)

        some_X = self.training_X[:3]
        some_y = ["one", "two", "three"]
        builder.add_samples(some_X, some_y)
        self._validate_samples(builder.spec, some_X, some_y)

        addl_X = self.training_X[3:6]
        addl_y = ["four", "five", "six"]
        builder.add_samples(addl_X, addl_y)
        self._validate_samples(builder.spec, self.training_X[0:6], some_y + addl_y)

    def test_add_samples_invalid_label_types(self):
        builder_int_labels = self.create_builder(default_class_label=42)
        self.assertIsNotNone(builder_int_labels)

        some_X = self.training_X[:3]
        invalid_int_y = [0, "one", 2]
        with self.assertRaises(TypeError):
            builder_int_labels.add_samples(some_X, invalid_int_y)

        builder_string_labels = self.create_builder(default_class_label="default")
        self.assertIsNotNone(builder_string_labels)

        invalid_string_y = ["zero", "one", 2]
        with self.assertRaises(TypeError):
            builder_string_labels.add_samples(some_X, invalid_string_y)

    @unittest.skipUnless(_is_macos(), "Only supported on MacOS platform.")
    def test_can_init_and_save_model_from_builder_with_updated_spec(self):
        builder = KNearestNeighborsClassifierBuilder(
            input_name="input",
            output_name="output",
            number_of_dimensions=10,
            default_class_label="defaultLabel",
            k=3,
            weighting_scheme="inverse_distance",
            index_type="kd_tree",
            leaf_size=50,
        )
        builder.author = "CoreML Team"
        builder.license = "MIT"
        builder.description = "test_builder_with_validation"

        # Save the updated spec
        coreml_model = MLModel(builder.spec)
        self.assertIsNotNone(coreml_model)
        coreml_model_path = "/tmp/__test_builder_with_validation.mlmodel"

        try:
            coreml_model.save(coreml_model_path)
            self.assertTrue(os.path.isfile(coreml_model_path))
        finally:
            self._delete_mlmodel_and_mlmodelc(coreml_model_path)

    @unittest.skipUnless(_is_macos(), "Only supported on MacOS platform.")
    def test_can_init_and_save_model_from_builder_default_parameters(self):
        builder = KNearestNeighborsClassifierBuilder(
            input_name="input",
            output_name="output",
            number_of_dimensions=4,
            default_class_label="defaultLabel",
        )

        # Save the updated spec
        coreml_model = MLModel(builder.spec)
        self.assertIsNotNone(coreml_model)
        coreml_model_path = "/tmp/__test_builder_with_validation.mlmodel"

        try:
            coreml_model.save(coreml_model_path)
            self.assertTrue(os.path.isfile(coreml_model_path))
        finally:
            self._delete_mlmodel_and_mlmodelc(coreml_model_path)

    def _validate_samples(self, spec, expected_X, expected_y):
        """Validate the float samples returned from the converted scikit KNeighborsClassifier"""
        num_dimensions = (
            spec.kNearestNeighborsClassifier.nearestNeighborsIndex.numberOfDimensions
        )
        for index, sample in enumerate(
            spec.kNearestNeighborsClassifier.nearestNeighborsIndex.floatSamples
        ):
            for dim in range(0, num_dimensions):
                self.assertAlmostEqual(
                    sample.vector[dim], expected_X[index][dim], places=6
                )

        if spec.kNearestNeighborsClassifier.HasField("int64ClassLabels"):
            for index, label in enumerate(
                spec.kNearestNeighborsClassifier.int64ClassLabels.vector
            ):
                self.assertEqual(label, expected_y[index])

        elif spec.kNearestNeighborsClassifier.HasField("stringClassLabels"):
            for index, label in enumerate(
                spec.kNearestNeighborsClassifier.stringClassLabels.vector
            ):
                self.assertEqual(label, expected_y[index])

    @staticmethod
    def _delete_mlmodel_and_mlmodelc(path_to_mlmodel):
        """Delete the .mlmodel and .mlmodelc for the given .mlmodel."""
        if os.path.exists(path_to_mlmodel):
            os.remove(path_to_mlmodel)
        path_to_mlmodelc = "{}c".format(path_to_mlmodel)
        if os.path.exists(path_to_mlmodelc):
            shutil.rmtree(path_to_mlmodelc)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import os
from ...models import (
    _MLMODEL_FULL_PRECISION,
    _MLMODEL_HALF_PRECISION,
    _VALID_MLMODEL_PRECISION_TYPES,
)


def convert(
    model,
    image_input_names=[],
    is_bgr=False,
    red_bias=0.0,
    blue_bias=0.0,
    green_bias=0.0,
    gray_bias=0.0,
    image_scale=1.0,
    class_labels=None,
    predicted_feature_name=None,
    model_precision=_MLMODEL_FULL_PRECISION,
):
    """
    Convert a Caffe model to the Core ML format.

    Parameters
    ----------
    model: str | (str, str) | (str, str, str) | (str, str, dict)

        A trained Caffe neural network model which can be represented as:

        - Path on disk to a trained Caffe model (``.caffemodel``).
        - A tuple of two paths. The first path is the path to the ``.caffemodel``
          file. The second is the path to the ``deploy.prototxt``.
        - A tuple of three paths. The first path is the path to the
          trained ``.caffemodel`` file. The second is the path to the
          ``deploy.prototxt``. The third is a path to the mean image binary, data
          which is subtracted from the input image as a preprocessing step.
        - A tuple of two paths to ``.caffemodel`` and ``.prototxt`` and a ``dict`` with
          image input names as keys and paths to mean image ``binaryprotos`` as values.
          The keys should be the same as the input names provided via the argument
          ``image_input_name``.

    image_input_names: [str] | str
        The name(s) of the input blob(s) in the Caffe model that can be treated
        as images by Core ML. All other inputs are treated as MultiArrays (N-D
        Arrays) by Core ML.

    is_bgr: bool | dict()
        Flag indicating the channel order the model internally uses to represent
        color images. Set to True if the internal channel order is BGR,
        otherwise it will be assumed RGB. This flag is applicable only if
        ``image_input_names`` is specified. To specify a different value for each
        image input, provide a dictionary with input names as keys.
        Note that this flag is about the models internal channel order.
        An input image can be passed to the model in any color pixel layout
        containing red, green and blue values (e.g. ``32BGRA`` or ``32ARGB``). This flag
        determines how those pixel values get mapped to the internal multiarray
        representation.

    red_bias: float | dict()
        Bias value to be added to the red channel of the input image. Defaults to ``0.0``.
        Applicable only if ``image_input_names`` is specified. To specify different
        values for each image input, provide a dictionary with input names as keys.

    blue_bias: float | dict()
        Bias value to be added to the the blue channel of the input image.
        Defaults to ``0.0``.
        Applicable only if ``image_input_names`` is specified.
        To specify different values for each image input, provide a dictionary
        with input names as keys.

    green_bias: float | dict()
        Bias value to be added to the green channel of the input image.
        Defaults to ``0.0``.
        Applicable only if ``image_input_names`` is specified.
        To specify different values for each image input provide a dictionary
        with input names as keys.

    gray_bias: float | dict()
        Bias value to be added to the input image (in grayscale). Defaults to ``0.0``.
        Applicable only if ``image_input_names`` is specified.
        To specify different values for each image input provide a dictionary
        with input names as keys.

    image_scale: float | dict()
        Value by which the input images will be scaled before bias is added and
        Core ML model makes a prediction. Defaults to ``1.0``.
        Applicable only if ``image_input_names`` is specified.
        To specify different values for each image input provide a dictionary
        with input names as keys.

    class_labels: str
        Filepath where classes are parsed as a list of newline separated
        strings. Class labels map the index of the output of a neural network
        to labels in a classifier. Provide this argument to get a model of
        type classifier.

    predicted_feature_name: str
        Name of the output feature for the class labels exposed in the Core ML
        model (applies to classifiers only). Defaults to ``classLabel``.

    model_precision: str
        Precision at which model will be saved. Currently full precision
        (float) and half precision (float16) models are supported. Defaults to
        ``_MLMODEL_FULL_PRECISION`` (full precision).

    Returns
    -------
    model: MLModel
        Model in the Core ML format.

    Examples
    --------
    .. sourcecode:: python

		# Convert it with default input and output names
   		>>> import coremltools
		>>> coreml_model = coremltools.converters.caffe.convert('my_caffe_model.caffemodel')

		# Saving the Core ML model to a file.
		>>> coreml_model.save('my_model.mlmodel')
		
        
    Sometimes, critical information in the Caffe converter is missing from the
    ``.caffemodel`` file. This information is present in the ``deploy.prototxt`` file.
    You can provide us with both files in the conversion process.

    .. sourcecode:: python

		>>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel', 'my_deploy.prototxt'))
        
        
    Some models (such as Resnet-50) also require a mean image file, which is
    subtracted from the input image before passing through the network. This
    file can also be provided during conversion:

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel',
        ...                 'my_deploy.prototxt', 'mean_image.binaryproto'), image_input_names = 'image_input')

        # Multiple mean images for preprocessing
        >>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel',
        ...                 'my_deploy.prototxt', {'image1': 'mean_image1.binaryproto', 'image2': 'mean_image2.binaryproto'}),
        ...                     image_input_names = ['image1', 'image2'])

        # Multiple image inputs and bias/scale values
        >>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel', 'my_deploy.prototxt'),
        ...                     red_bias = {'image1': -100, 'image2': -110},
        ...                     green_bias = {'image1': -90, 'image2': -125},
        ...                     blue_bias = {'image1': -105, 'image2': -120},
        ...                     image_input_names = ['image1', 'image2'])
        
        
    Input and output names used in the interface of the converted Core ML
    model are inferred from the ``.prototxt`` file, which contains a
    description of the network architecture. Input names are read from the
    input layer definition in the ``.prototxt``. By default, they are of type
    MultiArray. Argument ``"image_input_names"`` can be used to assign image
    type to specific inputs. All the blobs that are "dangling" -- which do not
    feed as input to any other layer -- are taken as outputs. The
    ``.prototxt`` file can be modified to specify custom input and output
    names.

    The converted Core ML model is of type classifier when the argument
    ``"class_labels"`` is specified.

    Advanced usage with custom classifiers, and images:

    .. sourcecode:: python

		# Mark some inputs as Images
		>>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel', 'my_caffe_model.prototxt'),
		...                   image_input_names = 'my_image_input')

		# Export as a classifier with classes from a file
		>>> coreml_model = coremltools.converters.caffe.convert(('my_caffe_model.caffemodel', 'my_caffe_model.prototxt'),
		...         image_input_names = 'my_image_input', class_labels = 'labels.txt')
		
		
    Sometimes the converter might return a message about not able to infer
    input data dimensions. This happens when the input size information is
    absent from the ``deploy.prototxt`` file. This can be easily provided by
    editing the ``.prototxt`` in a text editor. Simply add a snippet in the
    beginning, similar to the following, for each of the inputs to the model:

    .. code-block:: bash

        input: "my_image_input"
        input_dim: 1
        input_dim: 3
        input_dim: 227
        input_dim: 227
        
        
    In this example we have specified an input with dimensions ``(1,3,227,227)``, using
    Caffe's convention, in the order (batch, channel, height, width). Input
    name string (``"my_image_input"``) must also match the name of the input (or
    "bottom", as inputs are known in Caffe) of the first layer in the
    ``.prototxt``.

    """
    from ...models import MLModel
    from ...models.utils import _convert_neural_network_weights_to_fp16

    if model_precision not in _VALID_MLMODEL_PRECISION_TYPES:
        raise RuntimeError("Model precision {} is not valid".format(model_precision))

    import tempfile

    model_path = tempfile.mktemp()
    _export(
        model_path,
        model,
        image_input_names,
        is_bgr,
        red_bias,
        blue_bias,
        green_bias,
        gray_bias,
        image_scale,
        class_labels,
        predicted_feature_name,
    )
    model = MLModel(model_path)
    try:
        os.remove(model_path)
    except OSError:
        pass

    if model_precision == _MLMODEL_HALF_PRECISION and model is not None:
        model = _convert_neural_network_weights_to_fp16(model)

    return model


def _export(
    filename,
    model,
    image_input_names=[],
    is_bgr=False,
    red_bias=0.0,
    blue_bias=0.0,
    green_bias=0.0,
    gray_bias=0.0,
    image_scale=1.0,
    class_labels=None,
    predicted_feature_name=None,
):
    from ... import libcaffeconverter

    if isinstance(model, str):
        src_model_path = model
        prototxt_path = u""
        binaryproto_path = dict()
    elif isinstance(model, tuple):
        if len(model) == 3:
            src_model_path, prototxt_path, binaryproto_path = model
        else:
            src_model_path, prototxt_path = model
            binaryproto_path = dict()

    if isinstance(image_input_names, str):
        image_input_names = [image_input_names]
    if predicted_feature_name is None:
        predicted_feature_name = u"classLabel"
    if class_labels is None:
        class_labels = u""

    if binaryproto_path:
        if not image_input_names:
            raise RuntimeError(
                "'image_input_names' must be provided when a mean image binaryproto path is specified. "
            )

    if isinstance(binaryproto_path, str):
        binaryproto_paths = dict()
        binaryproto_paths[image_input_names[0]] = binaryproto_path
    elif isinstance(binaryproto_path, dict):
        binaryproto_paths = binaryproto_path
    else:
        raise RuntimeError(
            "Mean image binaryproto path must be a string or a dictionary of inputs names and paths. "
        )

    if not isinstance(is_bgr, dict):
        is_bgr = dict.fromkeys(image_input_names, is_bgr)
    if not isinstance(red_bias, dict):
        red_bias = dict.fromkeys(image_input_names, red_bias)
    if not isinstance(blue_bias, dict):
        blue_bias = dict.fromkeys(image_input_names, blue_bias)
    if not isinstance(green_bias, dict):
        green_bias = dict.fromkeys(image_input_names, green_bias)
    if not isinstance(gray_bias, dict):
        gray_bias = dict.fromkeys(image_input_names, gray_bias)
    if not isinstance(image_scale, dict):
        image_scale = dict.fromkeys(image_input_names, image_scale)

    libcaffeconverter._convert_to_file(
        src_model_path,
        filename,
        binaryproto_paths,
        set(image_input_names),
        is_bgr,
        red_bias,
        blue_bias,
        green_bias,
        gray_bias,
        image_scale,
        prototxt_path,
        class_labels,
        predicted_feature_name,
    )



"""Module containing unit tests for verifying various quantization."""
import os
import shutil
import tempfile
import unittest
import numpy as np
import pytest

import coremltools
import coremltools.models.datatypes as datatypes
from coremltools.models import neural_network
import coremltools.models.neural_network.quantization_utils as quantization_utils
from coremltools.models.neural_network.quantization_utils import (
    activate_int8_int8_matrix_multiplications,
    MatrixMultiplyLayerSelector,
    _quantize_spec_weights,
)

from coremltools._deps import _HAS_KERAS2_TF
from coremltools.models import (
    _MLMODEL_FULL_PRECISION,
    _QUANTIZATION_MODE_LINEAR_QUANTIZATION,
    _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS,
    _QUANTIZATION_MODE_CUSTOM_LOOKUP_TABLE,
)


@unittest.skipIf(
    not coremltools.utils._is_macos() or coremltools.utils._macos_version() < (10, 14),
    "Missing macOS 10.14+. Skipping tests.",
)
@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class QuantizationNumericalCorrectnessTests(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        self.qbits = 8  # n-bit quantization for tests
        self.qmode = _QUANTIZATION_MODE_LINEAR_QUANTIZATION
        self.custom_lut = None
        from .test_keras2_numeric import KerasBasicNumericCorrectnessTest

        self.keras_tester = KerasBasicNumericCorrectnessTest()
        self.keras_tester._test_model = self._test_model

    def _run_quantized_test(self, input_, full_precision_model, quantized_model, delta):
        # Output from both models should be the same
        full_output = full_precision_model.predict(input_)
        quantized_output = quantized_model.predict(input_)
        self.assertEqual(full_output.keys(), quantized_output.keys())

        for key in full_output.keys():
            full_output_flatten = full_output[key].flatten()
            quantized_output_flatten = quantized_output[key].flatten()

            self.assertTrue(len(full_output_flatten) == len(quantized_output_flatten))

            norm_factor = np.maximum(full_output_flatten, quantized_output_flatten)
            norm_factor = np.maximum(norm_factor, 1.0)
            f_out = full_output_flatten / norm_factor
            q_out = quantized_output_flatten / norm_factor

            for idx, full_value in enumerate(f_out):
                quantized_value = q_out[idx]
                self.assertAlmostEqual(full_value, quantized_value, delta=delta)

    def _test_model(
        self,
        model,
        num_samples=1,
        mode="random",
        delta=1e-2,
        model_dir=None,
        transpose_keras_result=True,
        one_dim_seq_flags=None,
        model_precision=_MLMODEL_FULL_PRECISION,
    ):
        # Get the model path
        use_tmp_folder = False
        if model_dir is None:
            use_tmp_folder = True
            model_dir = tempfile.mkdtemp()
        _ = os.path.join(model_dir, "keras.mlmodel")

        # Get converted coreml model and sample input
        (
            input_names,
            output_names,
            _,
            coreml_input,
        ) = self.keras_tester._get_coreml_model_params_and_test_input(
            model, mode, one_dim_seq_flags
        )
        from .test_keras2_numeric import _get_coreml_model

        coreml_model = _get_coreml_model(
            model, input_names, output_names, model_precision=model_precision
        )

        # Now we quantize the model and dequantize it. We then use this model
        # as our full precision model since quantizing this model again will
        # result in 0 quantization error.

        coreml_spec = coreml_model.get_spec()
        quantization_utils._quantize_spec_weights(
            spec=coreml_spec,
            nbits=self.qbits,
            quantization_mode=self.qmode,
            lut_function=self.custom_lut,
        )

        # De-quantize model
        quantization_utils._dequantize_nn_spec(spec=coreml_spec.neuralNetwork)
        full_precision_model_spec = coreml_spec

        # Quantize model from another copy
        quantized_model_spec = quantization_utils._quantize_spec_weights(
            spec=coreml_model.get_spec(),
            nbits=self.qbits,
            quantization_mode=self.qmode,
            lut_function=self.custom_lut,
        )

        full_precision_model = coremltools.models.MLModel(full_precision_model_spec)
        quantized_model = coremltools.models.MLModel(quantized_model_spec)
        self._run_quantized_test(
            coreml_input, full_precision_model, quantized_model, delta
        )

        # Clean up after ourselves
        if use_tmp_folder and os.path.exists(model_dir):
            shutil.rmtree(model_dir)

    def test_quantized_tiny_inner_product(self):
        self.keras_tester.test_tiny_inner_product()

    def test_quantized_conv_batchnorm_random(self):
        self.keras_tester.test_conv_batchnorm_random()

    def test_quantized_conv_batchnorm_no_gamma_no_beta(self):
        self.keras_tester.test_conv_batchnorm_no_gamma_no_beta()

    def test_quantized_tiny_deconv_random(self):
        self.keras_tester.test_tiny_deconv_random()

    def test_quantized_tiny_deconv_random_same_padding(self):
        self.keras_tester.test_tiny_deconv_random_same_padding()

    def test_quantized_tiny_depthwise_conv_valid_pad(self):
        self.keras_tester.test_tiny_depthwise_conv_valid_pad()

    def test_quantized_tiny_separable_conv_valid_depth_multiplier(self):
        self.keras_tester.test_tiny_separable_conv_valid_depth_multiplier()

    def test_quantized_max_pooling_no_overlap(self):
        self.keras_tester.test_max_pooling_no_overlap()

    def test_quantized_dense_softmax(self):
        self.keras_tester.test_dense_softmax()

    def test_quantized_housenet_random(self):
        self.keras_tester.test_housenet_random()

    def test_quantized_large_input_length_conv1d_same_random(self):
        self.keras_tester.test_large_input_length_conv1d_same_random()

    def test_quantized_conv_dense(self):
        self.keras_tester.test_conv_dense()

    def test_quantized_tiny_conv_crop_1d_random(self):
        self.keras_tester.test_tiny_conv_crop_1d_random()

    def test_quantized_embedding(self):
        self.keras_tester.test_embedding()

    def test_quantized_tiny_conv_elu_random(self):
        self.keras_tester.test_tiny_conv_elu_random()

    def test_quantized_tiny_concat_random(self):
        self.keras_tester.test_tiny_concat_random()

    def test_quantized_tiny_dense_tanh_fused_random(self):
        self.keras_tester.test_tiny_dense_tanh_fused_random()

    def test_quantized_conv1d_flatten(self):
        # Softmax after quantization appears to have a bigger error margin
        self.keras_tester.test_conv1d_flatten(delta=2e-2)

    def test_quantized_tiny_conv_dropout_random(self):
        self.keras_tester.test_tiny_conv_dropout_random()

    def test_quantized_tiny_mul_random(self):
        self.keras_tester.test_tiny_mul_random()

    def test_quantized_tiny_conv_thresholded_relu_random(self):
        self.keras_tester.test_tiny_conv_thresholded_relu_random()

    def test_quantized_tiny_seq2seq_rnn_random(self):
        self.keras_tester.test_tiny_seq2seq_rnn_random()

    def test_quantized_rnn_seq(self):
        self.keras_tester.test_rnn_seq()

    def test_quantized_medium_no_sequence_simple_rnn_random(self):
        self.keras_tester.test_medium_no_sequence_simple_rnn_random()

    def test_quantized_tiny_no_sequence_lstm_zeros(self):
        self.keras_tester.test_tiny_no_sequence_lstm_zeros()

    def test_quantized_tiny_no_sequence_lstm_ones(self):
        self.keras_tester.test_tiny_no_sequence_lstm_ones()

    def test_quantized_lstm_seq(self):
        self.keras_tester.test_lstm_seq()

    def test_quantized_medium_no_sequence_lstm_random(self):
        self.keras_tester.test_medium_no_sequence_lstm_random()

    def test_quantized_tiny_no_sequence_gru_random(self):
        self.keras_tester.test_tiny_no_sequence_gru_random()

    def test_quantized_gru_seq_backwards(self):
        self.keras_tester.test_gru_seq_backwards()

    def test_quantized_tiny_no_sequence_bidir_random(self):
        self.keras_tester.test_tiny_no_sequence_bidir_random()

    def test_quantized_tiny_no_sequence_bidir_random_gpu(self):
        self.keras_tester.test_tiny_no_sequence_bidir_random_gpu()

    def test_quantized_small_no_sequence_bidir_random(self):
        self.keras_tester.test_small_no_sequence_bidir_random()

    def test_quantized_medium_no_sequence_bidir_random(self):
        self.keras_tester.test_medium_no_sequence_bidir_random()

    def test_quantized_medium_bidir_random_return_seq_false(self):
        self.keras_tester.test_medium_bidir_random_return_seq_false()

    def test_quantized_tiny_sequence_lstm(self):
        self.keras_tester.test_tiny_sequence_lstm()

    def test_quantized__lstm_td(self):
        self.keras_tester.test_lstm_td()

    def test_quantized_large_channel_gpu(self):
        self.keras_tester.test_large_channel_gpu()

    def test_quantized_tiny_seq2seq_rnn_random(self):
        self.keras_tester.test_tiny_seq2seq_rnn_random()

    def test_quantized_lstm_seq_backwards(self):
        self.keras_tester.test_lstm_seq_backwards()


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class SevenBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(SevenBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 7


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class SixBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(SixBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 6


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class FiveBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(FiveBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 5


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class FourBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(FourBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 4


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class ThreeBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(ThreeBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 3


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class TwoBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(TwoBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 2


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class OneBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(OneBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 1


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class LUTQuantizationNumericalCorrectnessTests(QuantizationNumericalCorrectnessTests):
    def setUp(self):
        super(LUTQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 8
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS

    def test_quantized_custom_lut(self):
        pass


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class LUTSevenBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTSevenBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 7
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class LUTSixBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTSixBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 6
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class LUTFiveBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTFiveBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 5
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class LUTFourBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTFourBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 4
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class LUTThreeBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTThreeBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 3
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
@pytest.mark.slow
class LUTTwoBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTTwoBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 2
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class LUTOneBitQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTOneBitQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 1
        self.qmode = _QUANTIZATION_MODE_LOOKUP_TABLE_KMEANS


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class LUTCustomQuantizationNumericalCorrectnessTests(
    QuantizationNumericalCorrectnessTests
):
    def setUp(self):
        super(LUTCustomQuantizationNumericalCorrectnessTests, self).setUp()
        self.qbits = 8
        self.qmode = _QUANTIZATION_MODE_CUSTOM_LOOKUP_TABLE
        self.custom_lut = quantization_utils._get_linear_lookup_table_and_weight


from coremltools.converters import keras as keras_converter


@unittest.skipIf(
    not coremltools.utils._is_macos() or coremltools.utils._macos_version() < (10, 14),
    "Missing macOS 10.14+. Skipping tests.",
)
@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras2
class AdvancedQuantizationNumericalCorrectnessTests(unittest.TestCase):
    """ Quantization tests for advanced settings
    """

    def test_8bit_symmetric_and_skips(self):
        from keras.models import Sequential
        from keras.layers import Conv2D

        def stable_rel_error(x, ref):
            err = x - ref
            denom = np.maximum(np.abs(ref), np.ones_like(ref))
            return np.abs(err) / denom

        np.random.seed(1988)
        input_dim = 16
        num_kernels, kernel_height, kernel_width, input_channels = 64, 3, 3, 32

        # Define a model
        model = Sequential()
        model.add(
            Conv2D(
                input_shape=(input_dim, input_dim, input_channels),
                filters=num_kernels,
                kernel_size=(kernel_height, kernel_width),
            )
        )

        # Set some random weights
        weight, bias = model.layers[0].get_weights()
        num_filters = weight.shape[-1]
        filter_shape = weight.shape[:-1]

        new_weight = np.stack(
            [4.0 * np.random.rand(*filter_shape) - 2 for i in range(num_filters)],
            axis=-1,
        )
        model.layers[0].set_weights([new_weight, bias])

        mlmodel = keras_converter.convert(model, ["data"], ["output_0"])
        selector = quantization_utils.AdvancedQuantizedLayerSelector(
            skip_layer_types=["batchnorm", "bias", "depthwiseConv"],
            minimum_conv_kernel_channels=4,
            minimum_conv_weight_count=4096,
        )

        q_mlmodel = quantization_utils.quantize_weights(mlmodel, 8, selector=selector)

        input_shape = (1, 1, input_channels, input_dim, input_dim)
        input_val = 2 * np.random.rand(*input_shape) - 1

        coreml_input = {"data": input_val}
        coreml_output = mlmodel.predict(coreml_input)
        q_coreml_output = q_mlmodel.predict(coreml_input)

        val = coreml_output["output_0"]
        q_val = q_coreml_output["output_0"]
        rel_err = stable_rel_error(q_val, val)
        max_rel_err, mean_rel_err = np.max(rel_err), np.mean(rel_err)
        self.assertTrue(max_rel_err < 0.25)
        self.assertTrue(max_rel_err > 0.01)
        self.assertTrue(mean_rel_err < 0.02)


@unittest.skipIf(
    not coremltools.utils._is_macos() or coremltools.utils._macos_version() < (10, 16),
    "Missing macOS 10.16+. Skipping tests.",
)
class DynamicQuantizedInt8Int8MatMul(unittest.TestCase):
    """
    Quantization tests for dynamic Int8 - Int8 matrix multiplications
    """

    def initialize(self):
        np.random.seed(1988)
        self.Cout, self.Cin = 16, 32
        self.W = np.random.rand(self.Cout, self.Cin) * 20.0 - 10.0
        self.b = np.random.rand(self.Cout) * 20.0 - 10.0
        self.input_shape = (5, self.Cin)
        input_features = [("data", datatypes.Array(*self.input_shape))]
        output_features = [("output", None)]
        self.builder = neural_network.NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        self.selector = MatrixMultiplyLayerSelector()

    def _test_predictions(
        self, np_preds, coreml_preds, SNR=30, PSNR=40,
    ):

        np_preds = np_preds.flatten()
        coreml_preds = coreml_preds.flatten()

        noise = np_preds - coreml_preds
        noise_var = np.sum(noise ** 2) / len(noise) + 1e-7
        signal_energy = np.sum(np_preds ** 2) / len(np_preds)
        max_signal_energy = np.amax(np_preds ** 2)
        snr = 10 * np.log10(signal_energy / noise_var)
        psnr = 10 * np.log10(max_signal_energy / noise_var)
        self.assertGreaterEqual(snr, SNR)
        self.assertGreaterEqual(psnr, PSNR)

    def compare(self, specification_modified=True):
        x = np.random.rand(*self.input_shape)

        def _get_preds(spec):
            mlmodel = coremltools.models.MLModel(spec)
            return mlmodel.predict({"data": x}, useCPUOnly=True)["output"]

        preds = _get_preds(self.builder.spec)
        self.assertEqual(self.builder.spec.specificationVersion, 4)

        quantized_spec = activate_int8_int8_matrix_multiplications(
            self.builder.spec, self.selector
        )

        layer = self.builder.spec.neuralNetwork.layers[0]
        layer_type = layer.WhichOneof("layer")
        if layer_type == "innerProduct":
            matmul_layer = layer.innerProduct

        elif layer_type == "batchedMatmul":
            matmul_layer = layer.batchedMatmul
        wp = matmul_layer.weights

        if specification_modified:
            self.assertEqual(self.builder.spec.specificationVersion, 5)
            quant_preds = _get_preds(quantized_spec)
            self._test_predictions(preds, quant_preds, SNR=40)
            self.assertEqual(len(wp.floatValue), 0)
        else:
            self.assertEqual(self.builder.spec.specificationVersion, 4)
            quant_preds = _get_preds(quantized_spec)
            np.testing.assert_array_almost_equal(preds, quant_preds)
            self.assertGreater(len(wp.floatValue), 0)

    def test_single_batched_matmul_no_bias(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.compare()

    def test_single_batched_matmul_with_bias(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
            bias=self.b,
        )
        self.compare()

    def test_single_inner_product_no_bias(self):

        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=None,
            has_bias=False,
        )
        self.compare()

    def test_single_inner_product_with_bias(self):

        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.compare()

    def test_inner_product_min_input_channels_valid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.minimum_input_channels = 31
        self.compare()

    def test_batched_matmul_min_input_channels_valid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.minimum_input_channels = 32
        self.compare()

    def test_inner_product_min_input_channels_invalid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.minimum_input_channels = 33
        self.compare(specification_modified=False)

    def test_batched_matmul_min_input_channels_invalid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.minimum_input_channels = 33
        self.compare(specification_modified=False)

    def test_batched_matmul_max_input_channels_valid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.maximum_input_channels = 32
        self.compare()

    def test_inner_product_max_input_channels_valid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.maximum_input_channels = 33
        self.compare()

    def test_batched_matmul_max_input_channels_invalid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.maximum_input_channels = 31
        self.compare(specification_modified=False)

    def test_inner_product_max_input_channels_invalid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.maximum_input_channels = 30
        self.compare(specification_modified=False)

    def test_inner_product_min_output_channels_valid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.minimum_output_channels = 16
        self.compare()

    def test_batched_matmul_min_output_channels_valid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.minimum_output_channels = 16
        self.compare()

    def test_inner_product_min_output_channels_invalid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.minimum_output_channels = 17
        self.compare(specification_modified=False)

    def test_batched_matmul_min_output_channels_invalid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.minimum_output_channels = 17
        self.compare(specification_modified=False)

    def test_batched_matmul_max_output_channels_valid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.maximum_output_channels = 17
        self.compare()

    def test_inner_product_max_output_channels_valid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.maximum_output_channels = 16
        self.compare()

    def test_batched_matmul_max_output_channels_invalid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.maximum_output_channels = 14
        self.compare(specification_modified=False)

    def test_inner_product_max_output_channels_invalid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.maximum_output_channels = 15
        self.compare(specification_modified=False)

    def test_inner_product_min_weight_count_valid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.minimum_weight_count = 512
        self.compare()

    def test_batched_matmul_min_weight_count_invalid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.minimum_weight_count = 513
        self.compare(specification_modified=False)

    def test_inner_product_layer_names_invalid(self):
        self.initialize()
        self.builder.add_inner_product(
            name="ip",
            input_name="data",
            output_name="output",
            input_channels=self.Cin,
            output_channels=self.Cout,
            W=self.W,
            b=self.b,
            has_bias=True,
        )
        self.selector.include_layers_with_names = ["ip1", "ip2"]
        self.compare(specification_modified=False)

    def test_batched_matmul_layer_names_valid(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        self.selector.include_layers_with_names = ["bm1", "batched_matmul"]
        self.compare()

    def test_batched_matmul_8bit_weight_quantized(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        _quantize_spec_weights(
            self.builder.spec, 8, _QUANTIZATION_MODE_LINEAR_QUANTIZATION
        )
        self.compare()

    def test_batched_matmul_4bit_weight_quantized(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        _quantize_spec_weights(
            self.builder.spec, 4, _QUANTIZATION_MODE_LINEAR_QUANTIZATION
        )
        self.compare()

    def test_batched_matmul_2bit_weight_quantized(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        _quantize_spec_weights(
            self.builder.spec, 2, _QUANTIZATION_MODE_LINEAR_QUANTIZATION
        )
        self.compare()

    def test_batched_matmul_1bit_weight_quantized(self):

        self.initialize()
        self.builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="output",
            weight_matrix_rows=self.Cin,
            weight_matrix_columns=self.Cout,
            W=self.W,
        )
        _quantize_spec_weights(
            self.builder.spec, 1, _QUANTIZATION_MODE_LINEAR_QUANTIZATION
        )
        self.compare()


@unittest.skipIf(
    not coremltools.utils._is_macos() or coremltools.utils._macos_version() < (10, 15),
    "Missing macOS 10.15+. Skipping tests.",
)
class QuantizeWeightsAPI(unittest.TestCase):
    def test_embeddingND_quantize(self):
        input_features = [("data", datatypes.Array(10, 1))]
        output_features = [("output", None)]
        builder = neural_network.NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )

        builder.add_embedding_nd(
            name="embedding_nd",
            input_name="data",
            output_name="output",
            vocab_size=300,
            embedding_size=20,
            W=np.random.rand(20, 300),
        )

        spec = builder.spec
        model_fp32 = coremltools.models.MLModel(spec)
        self.assertEqual(
            len(spec.neuralNetwork.layers[0].embeddingND.weights.floatValue), 6000
        )

        # quantize to FP16
        model_fp16 = quantization_utils.quantize_weights(model_fp32, nbits=16)
        spec_fp16 = model_fp16.get_spec()
        self.assertEqual(
            len(spec_fp16.neuralNetwork.layers[0].embeddingND.weights.floatValue), 0
        )
        self.assertEqual(
            len(spec_fp16.neuralNetwork.layers[0].embeddingND.weights.float16Value),
            2 * 6000,
        )

        # quantize to uint8
        model_uint8 = quantization_utils.quantize_weights(model_fp32, nbits=8)
        spec_uint8 = model_uint8.get_spec()
        self.assertEqual(
            len(spec_uint8.neuralNetwork.layers[0].embeddingND.weights.floatValue), 0
        )
        self.assertEqual(
            len(spec_uint8.neuralNetwork.layers[0].embeddingND.weights.float16Value), 0
        )
        self.assertEqual(
            len(spec_uint8.neuralNetwork.layers[0].embeddingND.weights.rawValue), 6000
        )

        # quantize to uint5
        model_uint5 = quantization_utils.quantize_weights(model_fp32, nbits=5)
        spec_uint5 = model_uint5.get_spec()
        self.assertEqual(
            len(spec_uint5.neuralNetwork.layers[0].embeddingND.weights.floatValue), 0
        )
        self.assertEqual(
            len(spec_uint5.neuralNetwork.layers[0].embeddingND.weights.float16Value), 0
        )
        self.assertEqual(
            len(spec_uint5.neuralNetwork.layers[0].embeddingND.weights.rawValue), 3750
        )  # 3750 = 5*6000/8
